//-----------------------------------------------------------------------------
/*
Programmed by Olivier Sangala
http://olivier.sangala.free.fr
olivier_sangala@hotmail.com
*/
//-----------------------------------------------------------------------------

#define USE_TI89              // Compile for TI-89
#define USE_TI92PLUS          // Compile for TI-92 Plus
#define USE_V200              // Compile for V200

// #define OPTIMIZE_ROM_CALLS // Use ROM Call Optimization

#define MIN_AMS 100           // Compile for AMS 1.00 or higher

#define SAVE_SCREEN           // Save/Restore LCD Contents

#include <tigcclib.h>         // Include All Header Files
#include <extgraph.h>

#include "menulib.h"
#include "kalah.h"

//--------------------------------------------------------------------------

extern TMenu * Menu;
enum {ME_MAIN,ME_PAUSE,ME_OPTIONS};

//--------------------------------------------------------------------------


char * player_str[] = {"Human","CPU1","CPU2","CPU3"};
char * yesno_str[] = {"No","Yes"};
char * turn_str[] = {"South","North"};

char *Plane0;


TGame *Game;
TBoard*Board;
TPoint Positions[14];

char StatusMsg[32];

//--------------------------------------------------------------------------

unsigned short Hole_C0[] = {0x7FFE,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x7FFE};
unsigned long Cursor0[] = {0x06000000,0x09000000,0x09000000,0x09000000,0x09000000,0x09C00000,0x09380000,0x09260000,0x09250000,0xE9248000,0x98048000,0x88008000,0x48008000,0x28008000,0x20008000,0x10008000,0x10010000,0x08010000,0x08010000,0x04020000,0x04020000,0x07FE0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000};

//--------------------------------------------------------------------------


int Algo_Evaluate(void)
{
	
	int score=(Board->Holes[6]-Board->Holes[13]);
	
	return (Game->algo_turn==0)?score:-score;
	
}
//--------------------------------------------------------------------------

int Algo_AlphaBeta(int depth, int A, int B)
{
	
	int a;
	
	// terminal node
	
	if (depth == Game->depth_max || Board->win_state!=WS_NONE)
	return Algo_Evaluate();
	
	
	// depth=0 -> init best score
	
	if (depth==0)
	Game->best_score=-INFINITY;
	
	int AP=-INFINITY;
	int BP=INFINITY;
	
	// VALID MOVES LIST
	
	int pos_start=(Board->turn?7:0);
	
	int valid_moves_count=0;
	int valid_moves_list[6];
	
	for (a=0;a<6;a++)
	{
	
		int pos=pos_start+a;
		
		if (Board->Holes[pos])
		valid_moves_list[valid_moves_count++]=pos;
	
	}
	
	// LOOP FOR LEGAL MOVES
	
	int node_max=(Game->algo_turn==Board->turn);
	
	for (a=0;a<valid_moves_count;a++)
	{
		
		int pos=valid_moves_list[a];
		
		// MAKE MOVE
		
		PlayPos(pos);
		
		int score = 0;
		
		if (node_max)
		{
			score = Algo_AlphaBeta(depth+1,max(A,AP),B);
			AP = max(AP,score);
		}
		else
		{
			score = Algo_AlphaBeta(depth+1,A,min(B,BP));
			BP = min(BP,score);
		}
		
		// depth=0 -> save score and best play
		
		if (depth==0)
		{
			
			if (score > Game->best_score)
			{
					Game->best_score=score;
					Game->best_count=0;
					Game->best_list[Game->best_count++] = pos;
			}
			else if (score == Game->best_score)
			{
					Game->best_list[Game->best_count++] = pos;
			}
			
		}
		
		// UNDO
		
		UndoMove();
		
		// alpha beta cut off
		
		if (node_max)
		{
		if (AP>=B)
		return AP;
		}
		else
		{
		if (BP>=A)
		return BP;
		}
		
		
	}
	
		// returning value
	
		if (node_max)
		return AP;
		else
		return BP;
	
}
//--------------------------------------------------------------------------

int Algo_GetBestPos(void)
{
	
	// Message
	strcpy(StatusMsg,"Thinking...");
	PlaceCursor();
	RefreshScreen();
	
	// init settings
	
	Game->algo_turn = Board->turn;
	
	// DEPTH
	
	int depth_list[]={2,2,3,4};
	int type = Game->AI_Computers[Board->turn];
	Game->depth_max = depth_list[type];
	
	// ALPHA BETA ALGORITHM
	
	Game->algo_active = 1;
	Algo_AlphaBeta(0,-INFINITY, INFINITY);
	Game->algo_active = 0;
	
	
	// Message
	strcpy(StatusMsg,"");
	PlaceCursor();
	RefreshScreen();
	
	// return best move
	return Game->best_list[random(Game->best_count)];
	
}
//--------------------------------------------------------------------------

void Algo_ShowAlgoInfo(void)
{

	clrscr();
	
	printf("depth_max : %d \n",Game->depth_max);
	printf("best_score : %d \n",Game->best_score);
	printf("best_play : %d \n",Game->best_play);
	
	ngetchx();
	
}
//--------------------------------------------------------------------------

void DrawHole (int Hole)
{
	char buffer[16];
	
	TPoint Pos = Positions[Hole];
	int Counters = Board->Holes[Hole];
	
	Sprite16_OR(Pos.x,Pos.y,16,Hole_C0,Plane0);
	
	if (Counters)
	{
		sprintf(buffer,"%d",Counters);
		DrawStr(Pos.x+4,Pos.y+5,buffer,A_NORMAL);
	}
	
}
//--------------------------------------------------------------------------

void RefreshScreen(void)
{
	int a;
	char buffer[50];
	
	FontSetSys(0);
	PortSet(Plane0,239,127);
	ClrScr();
	
	// DrawHole
	for (a=0;a<14;a++)
	DrawHole(a);
	
	// Stats
	sprintf(buffer,"Player : %s",turn_str[Board->turn]);
	DrawStr(0,0,buffer,A_NORMAL);
	
	sprintf(buffer,"%s",player_str[Game->AI_Computers[Board->turn]]);
	DrawStr(0,6,buffer,A_NORMAL);
	
	sprintf(buffer,"Moves : %d ",Game->step);
	DrawStr(0,12,buffer,A_NORMAL);
	
	char WinMsg[32];
	
	if (Board->win_state==WS_WIN0)
	sprintf(WinMsg,"Player %s wins!",turn_str[0]);
	else if (Board->win_state==WS_WIN1)
	sprintf(WinMsg,"Player %s wins!",turn_str[1]);
	else if (Board->win_state==WS_EXAEQUO)
	sprintf(WinMsg,"Ex aequo!");
	else
	strcpy(WinMsg,"");
	
	// StatusMsg
	DrawStr(80,0,StatusMsg,A_NORMAL);
	// WinMsg
	DrawStr(80,6,WinMsg,A_NORMAL);
	
	// DrawCursor
	DrawCursor();
	
	FontSetSys(1);
	PortRestore();
	FastCopyScreen(Plane0,LCD_MEM);
	
}
//--------------------------------------------------------------------------

void DrawCursor(void)
{
	char buffer[50];
	
	int x = Game->MousePos.x-CURSORFINGER_X+CURSORTILE_X;
	int y = Game->MousePos.y-CURSORFINGER_Y+CURSORTILE_Y;
	
	Sprite32_OR(x,y,CURSOR_HEIGHT,Cursor0,Plane0);
	
	if (Game->hand>0)
	{
		sprintf(buffer,"%d",Game->hand);
		DrawStr(x+CURSORHAND_X,y+CURSORHAND_Y,buffer,A_NORMAL);
	}
	
	
}
//--------------------------------------------------------------------------

void PlaceCursor(void)
{
	Game->MousePos = Positions[Board->position];
}

//--------------------------------------------------------------------------

void MoveHoleCounter(int HoleSrc, int HoleDest, int Count)
{
	
	if (!Game->algo_active)
	{
		
		if (HoleSrc!=HAND_HOLE)
		MouseMove(Positions[HoleSrc]);
		
		if (HoleDest!=HAND_HOLE)
		MouseMove(Positions[HoleDest]);
		
		RefreshScreen();
		
	}
	
	// HoleSrc
	
	if (HoleSrc==HAND_HOLE)
	Game->hand-=Count;
	else
	Board->Holes[HoleSrc]-=Count;
	
	// HoleDest
	
	if (HoleDest==HAND_HOLE)
	Game->hand+=Count;
	else
	Board->Holes[HoleDest]+=Count;
	
	
	
}
//--------------------------------------------------------------------------

void PlayPos ( int Position)
{
	int HoleNum;
	
	// IF NO COUNTERS -> RETURN
	if (!Board->Holes[Position])
	return;
	
	
	// HISTORY
	

	Game->Boards[Game->step+1]=Game->Boards[Game->step];
	
	Game->step++;
	Board++;
	
	// move counters
	
	MoveHoleCounter(Position,HAND_HOLE,Board->Holes[Position]);
	
	HoleNum = Position;
	int LastHole_Count=0;
	
	while (Game->hand)
	{
		
		
		// CORRECT HOLE LOOP
		
		next_hole_lbl:
		
		HoleNum++;
		HoleNum%=14;
		
		// skip_opp_kalahah;
		if (Game->skip_opp_kalahah)
		{
			if (((Board->turn==0 && HoleNum==13) || (Board->turn==1 && HoleNum==6)))
			goto next_hole_lbl;
		}
		
		// skip_initial_hole
		if (Game->skip_initial_hole)
		{
			if (HoleNum==Position)
			goto next_hole_lbl;
		}
		
		LastHole_Count=Board->Holes[HoleNum];
		MoveHoleCounter(HAND_HOLE,HoleNum,1);
		
	}
	
	
	
	int HoleTurn=(HoleNum>6?1:0);
	
	
	// IF PLACE COUNTER IN EMPTY HOLE
	if (LastHole_Count==0 && HoleTurn==Board->turn && (HoleNum!=6 && HoleNum!=13))
	{
		
		
		// null_captures;
		if (!(Board->Holes[HoleNum]==0 && !Game->null_captures))
		{
			
			MoveHoleCounter(HoleNum,HAND_HOLE,Board->Holes[HoleNum]);
			
			int Score_HoleNum=(Board->turn?13:6);
			MoveHoleCounter(HAND_HOLE,Score_HoleNum,Game->hand);
			
			int Opposite_HoleNum = 12-HoleNum;
			
			// IF COUNTERS IN OPPOSITE SIDE
			if (Board->Holes[Opposite_HoleNum]>0)
			{
				
				MoveHoleCounter(Opposite_HoleNum,HAND_HOLE,Board->Holes[Opposite_HoleNum]);
				
				int OppositeScore_HoleNum=(Board->turn?13:6);
				MoveHoleCounter(HAND_HOLE,OppositeScore_HoleNum,Game->hand);
				
				
			}
			
		}
		
		
	}
	
	
	
	// LAST COUNTER IN KALAHAH -> REPLAY
	if (!((HoleNum==6 && Board->turn==0) || (HoleNum==13 && Board->turn==1 )))
	{
		Board->turn^=1;
		BeginTurn();
	}
	
	CheckForEndGame();
	
	
}
//--------------------------------------------------------------------------

void CheckForEndGame(void)
{
	int a;
	
	int Counters[2];
	
	Board->win_state = WS_NONE;
	
	// CheckForEndGame
	
	Counters[0] = 0;
	for (a=0;a<=5;a++)
	Counters[0]+=Board->Holes[a];
	
	Counters[1] = 0;
	for (a=7;a<=12;a++)
	Counters[1]+=Board->Holes[a];
	
	// IF GAME OVER
	
	if (Counters[0]==0 || Counters[1]==0)
	{
		
		// COLLECT REMAINING COUNTERS
		
		int pos_collect=(Counters[0]?0:7);
		int count=(Counters[0]?Counters[0]:Counters[1]);
		
		for (a=0;a<6;a++)
		Board->Holes[pos_collect+a] = 0;
		
		Board->Holes[pos_collect+6]+=count;
		
		// win_state
		
		if (Board->Holes[6]>Board->Holes[13])
		Board->win_state = WS_WIN0;
		else if (Board->Holes[13]>Board->Holes[6])
		Board->win_state = WS_WIN1;
		else
		Board->win_state = WS_EXAEQUO;
		
		
	}
	
}
//--------------------------------------------------------------------------


void NewGame(void)
{
	int a;
	
	Game->step = 0;
	
	// BOARD POINTER
	Board = Game->Boards;
	
	for (a=0;a<14;a++)
	Board->Holes[a] = ((a==6 || a==13)?0:Game->counters);
	
	
	Board->turn = 0;
	Board->win_state = WS_NONE;
	
	
	Game->hand = 0;
	
	// MESSAGE
	strcpy(StatusMsg,"");
	
	BeginTurn();
	
}

//--------------------------------------------------------------------------


void MouseMove(TPoint Point)
{
	
	
	int Counter=0;
	
	RefreshScreen();
	
	for (;;)
	{
		if (Game->MousePos.x>Point.x)
		Game->MousePos.x--;
		else if (Game->MousePos.x<Point.x)
		Game->MousePos.x++;
		else if (Game->MousePos.y>Point.y)
		Game->MousePos.y--;
		else if (Game->MousePos.y<Point.y)
		Game->MousePos.y++;
		else
		break;
		
		if (!(Counter++&1))
		RefreshScreen();
		
	}
	
	RefreshScreen();
	
}
//--------------------------------------------------------------------------


void UndoMove(void)
{
	
	if (Game->step)
	{
		Board--;
		Game->step--;
	}
	
}
//--------------------------------------------------------------------------


void BeginTurn(void)
{
	Board->position = (Board->turn?10:3);
}
//--------------------------------------------------------------------------

// Main Function
void _main(void)
{
	int a;
	
	TGame GameA;
	Game=&GameA;
	
	TMenu MenuA;
	Menu = &MenuA;
	
	Game->algo_active = 0;
	
	if (!(Plane0=malloc(LCD_SIZE)))
	goto end_prog;
	
	
	// Positions
	
	for (a=0;a<6;a++)
	{
		Positions[a].x = POS_LEFT+20+20*a;
		Positions[a].y = POS_TOP+40;
		Positions[a+7].x = POS_LEFT+120-20*a;
		Positions[a+7].y = POS_TOP;
	}
	
	
	Positions[6].x = POS_LEFT+120;
	Positions[6].y = POS_TOP+20;
	Positions[13].x = POS_LEFT+20;
	Positions[13].y = POS_TOP+20;
	
	// Options
	
	Game->AI_Computers[0] = TP_HUMAN;
	Game->AI_Computers[1] = TP_CPU1;
	Game->counters = 4;
	Game->skip_initial_hole = 0;
	Game->null_captures = 1;
	Game->skip_opp_kalahah = 1;
	
	randomize();
	
	for (;;)
	{
		
		MenuNew2();
		MenuAddTitle("Kalah");
		MenuAddItem("Play");
		MenuAddItem("Options");
		MenuAddItem("About");
		MenuAddItem("Quit");
		
		MenuDo(ME_MAIN);
		
		if (!Menu->ok) break;
		
		if (Menu->Index==1)
		{
			
			NewGame();
			
			for (;;)
			{
				
				// AI_Computers
				if (Game->AI_Computers[Board->turn]!=TP_HUMAN && Board->win_state==WS_NONE)
				{
					
					if (kbhit())
					if (ngetchx()==KEY_ESC)
					{
						Game->AI_Computers[0]=TP_HUMAN;
						Game->AI_Computers[1]=TP_HUMAN;
						continue;
					}
					
					Board->position = Algo_GetBestPos();
					
					PlaceCursor();
					RefreshScreen();
					
					PlayPos(Board->position);
					continue;
				}
				
				PlaceCursor();
				RefreshScreen();
				
				int key = ngetchx();
				
				int pos_min=(Board->turn?7:0);
				int pos_max=(Board->turn?12:5);
				
				
				if ((key==KEY_LEFT && Board->turn==0)
				|| (key==KEY_RIGHT && Board->turn==1))
				{
					if (Board->position>pos_min)
					Board->position--;
					else
					Board->position = pos_max;
				}
				
				else if ((key==KEY_RIGHT && Board->turn==0)
				|| (key==KEY_LEFT && Board->turn==1))
				{
					if (Board->position<pos_max)
					Board->position++;
					else
					Board->position = pos_min;
				}
				
				else if (key==KEY_F1 && Board->win_state==WS_NONE)
				{
					PlayPos(Board->position);
				}
				
				else if (key==KEY_F2)
				{
					NewGame();
				}
				
				else if (key==KEY_F3)
				{
					Game->AI_Computers[Board->turn]=TP_CPU1;
				}
				
				/*
				else if (key==KEY_F4)
				{
					Algo_ShowAlgoInfo();
				}
				*/
				
				else if (key==KEY_F5 && Board->win_state==WS_NONE)
				{
					Board->position = Algo_GetBestPos();
				}
				
				else if (key==KEY_BACKSPACE)
				{
					UndoMove();
					Game->AI_Computers[Board->turn]=TP_HUMAN;
				}
				
				else if (key==KEY_APPS)
				{
					off();
				}
				
				else if (key==KEY_ESC)
				{
					MenuNew2();
					MenuAddTitle("Exit Game?");
					MenuAddItem("Yes");
					MenuAddItem("No");
					MenuDo(ME_PAUSE);
					
					if (Menu->ok && Menu->Index==1)
					break;
					
				}
				
			}
			
			
		}
		else if (Menu->Index==2)
		{
			
			char buffer[50];
			
			for (;;)
			{
				
				MenuNew2();
				MenuAddTitle("Options");
				sprintf(buffer,"Player %s : %s",turn_str[0],player_str[Game->AI_Computers[0]]);
				MenuAddItem(buffer);
				sprintf(buffer,"Player %s : %s",turn_str[1],player_str[Game->AI_Computers[1]]);
				MenuAddItem(buffer);
				sprintf(buffer,"Initial Counters : %d",Game->counters);
				MenuAddItem(buffer);
				sprintf(buffer,"Skip Initial Hole : %s",yesno_str[Game->skip_initial_hole]);
				MenuAddItem(buffer);
				sprintf(buffer,"Skip Opp. Kalahah : %s",yesno_str[Game->skip_opp_kalahah]);
				MenuAddItem(buffer);
				sprintf(buffer,"Null Captures : %s",yesno_str[Game->null_captures]);
				MenuAddItem(buffer);
				
				
				MenuDo(ME_OPTIONS);
				
				if (!Menu->ok) break;
				
				if (Menu->Index==1)
				{
					Game->AI_Computers[0]++;
					if (Game->AI_Computers[0]==4)
					Game->AI_Computers[0] = 0;
				}
				else if (Menu->Index==2)
				{
					Game->AI_Computers[1]++;
					if (Game->AI_Computers[1]==4)
					Game->AI_Computers[1] = 0;
				}
				else if (Menu->Index==3)
				{
					if (Game->counters==6)
					Game->counters = 2;
					else
					Game->counters++;
				}
				else if (Menu->Index==4)
				Game->skip_initial_hole^=1;
				else if (Menu->Index==5)
				Game->skip_opp_kalahah^=1;
				else if (Menu->Index==6)
				Game->null_captures^=1;
				
			}
		}
		else if (Menu->Index==3)
		{
			FontSetSys(0);
			clrscr();
			printf("About \n");
			printf("Kalah 1.5 \n");
			printf("Programmed by Olivier Sangala \n");
			printf("http://olivier.sangala.free.fr \n");
			printf("olivier_sangala@hotmail.com \n");
			FontSetSys(1);
			ngetchx();
			
		}
		else
		break;
		
	}
	
	end_prog:
	
	if (Plane0)
	free(Plane0);
	
}
